"""Command-line interface for running mepylome.

Usage:
    $ mepylome                     # Run mepylome with default settings
    $ mepylome -a /path/to/idats   # Specify analysis IDAT files directory
               -r /path/to/ref     # Specify reference IDAT directory
               -c 450k             # Specify CpG's to use
               -s                  # Add horizontal segmentation lines
    $ mepylome --help              # Show all parameters

"""

import argparse
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path


def get_app_version():
    """Retrieve the app version from the package metadata."""
    try:
        return version("mepylome")
    except PackageNotFoundError:
        return "unknown"


def print_welcome_message():
    """Prints ASCII art welcome message."""
    app_version = get_app_version()
    welcome_message = rf"""
                             _
                            | |
  _ __ ___   ___ _ __  _   _| | ___  _ __ ___   ___
 | '_ ` _ \ / _ \ '_ \| | | | |/ _ \| '_ ` _ \ / _ \
 | | | | | |  __/ |_) | |_| | | (_) | | | | | |  __/
 |_| |_| |_|\___| .__/ \__, |_|\___/|_| |_| |_|\___|
                | |     __/ |
                |_|    |___/


Version: {app_version}
Starting Methylation Analysis...
"""
    print(welcome_message)


def absolute_path(path):
    """Converts a relative path to an absolute path."""
    return Path(path).absolute()


def parse_args():
    """Parses command line arguments."""
    parser = argparse.ArgumentParser(
        description="Command-line interface for running mepylome."
    )
    parser.add_argument(
        "-a",
        "--analysis_dir",
        type=absolute_path,
        help="Directory that contains the IDAT files to be analyzed.",
    )
    parser.add_argument(
        "-t",
        "--test_dir",
        type=absolute_path,
        help=(
            "Directory for test files, including new cases for analysis or "
            "validation."
        ),
    )
    parser.add_argument(
        "-A", "--annotation", type=str, help="Path to the annotation file."
    )
    parser.add_argument(
        "-r",
        "--reference_dir",
        type=absolute_path,
        help="Directory that contains CNV neutral reference cases.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=absolute_path,
        help="Directory for output data.",
    )
    parser.add_argument(
        "-p",
        "--prep",
        type=str,
        help=(
            "Preparation method. Possible values are 'illumina', 'swan', "
            "and 'noob'."
        ),
    )
    parser.add_argument(
        "-c",
        "--cpgs",
        type=str,
        nargs="+",
        help=(
            "CpGs to use for UMAP. Possible values are 'all', '450k', "
            "'epic', or 'epicv2'."
        ),
    )
    parser.add_argument(
        "-n", "--n_cpgs", type=int, help="Number of CpGs to use for UMAP."
    )
    parser.add_argument(
        "-C",
        "--precalculate_cnv",
        action="store_true",
        help=(
            "Precalculate CNV data in advance. This will take 1-2 seconds "
            "per case and improve performance during run time."
        ),
    )
    parser.add_argument(
        "-l",
        "--no_load_full_betas",
        action="store_false",
        dest="load_full_betas",
        help=(
            "Prevent loading betas for all CpG's into memory. By default, "
            "betas are loaded for all CpG's to improve the speed of "
            "generating "
            "multiple UMAP plots. Use this option to avoid "
            "potential memory overflow if insufficient memory is available "
            "(3-4 MB per sample needed)."
        ),
    )
    parser.add_argument(
        "-O",
        "--overlap",
        action="store_true",
        help=(
            "Only select IDAT files in the analysis directory that are also "
            "present in the annotation."
        ),
    )
    parser.add_argument(
        "-S",
        "--cpg_selection",
        type=str,
        default="top",
        help=(
            "Method for selecting the number of CpGs. Either 'random' or "
            "'top', where 'top' selects the CpGs with the highest variation "
            "within the analysis directory."
        ),
    )
    parser.add_argument(
        "-H", "--host", type=str, help="Host for the web server."
    )
    parser.add_argument(
        "-P", "--port", type=int, help="Port for the web server."
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Use debug mode in dash."
    )
    parser.add_argument(
        "-s",
        "--do_seg",
        action="store_true",
        help=(
            "Add horizontal segmentation lines in CNV plot "
            "(adds 1-2 seconds per sample)."
        ),
    )
    parser.add_argument(
        "--tutorial",
        action="store_true",
        help=(
            "Downloads test IDAT files used in the tutorial and then launches "
            "the mepylome GUI session. Use this for a quick demonstration of "
            "how this package works."
        ),
    )

    return parser.parse_args()


def start_mepylome():
    """Entry point to start mepylome methylation analysis from command line."""
    args = parse_args()
    cli_args = {k: v for k, v in vars(args).items() if v is not None}

    print_welcome_message()

    if not cli_args["load_full_betas"] and cli_args["cpg_selection"] == "top":
        cli_args["cpg_selection"] = "random"
        msg = (
            "Warning: Since '-no_load_full_betas' is set, 'cpg_selection' "
            "has been automatically changed to 'random'.\n"
        )
        print(msg)

    from .methyl import MethylAnalysis

    if cli_args["tutorial"]:
        from mepylome.utils import setup_tutorial_files

        tutorial_dir = Path.home() / "mepylome" / "tutorial"
        cli_args["analysis_dir"] = tutorial_dir / "tutorial_analysis"
        cli_args["reference_dir"] = tutorial_dir / "tutorial_reference"
        if (
            not cli_args["analysis_dir"].exists()
            and not cli_args["reference_dir"].exists()
        ):
            "Download Tutorial IDAT files"
            setup_tutorial_files(
                cli_args["analysis_dir"], cli_args["reference_dir"]
            )
        cli_args["load_full_betas"] = True
        cli_args["cpgs"] = "epic"

    cli_args.pop("tutorial", None)

    methyl_analysis = MethylAnalysis(**cli_args)
    methyl_analysis.run_app(open_tab=True)



"""Unittest for IDAT parser."""

import unittest

import numpy as np

from mepylome.dtypes.idat import IdatParser
from mepylome.tests.helpers import TempIdatFile


class TestIdatParser(unittest.TestCase):
    """Unittest for IDAT parser."""

    def _check_idat_parser(self, test_data, gzipped):
        """Check the data consistency between writer and parser."""
        tmp_idat = TempIdatFile(test_data, gzipped)
        idat_file = IdatParser(tmp_idat.path)
        for key in tmp_idat.data:
            test_data = tmp_idat.data[key]
            parsed_data = getattr(idat_file, key)
            if isinstance(test_data, np.ndarray):
                self.assertTrue(
                    np.array_equal(test_data, parsed_data),
                    f"Arrays for key '{key}' do not match.",
                )
            else:
                self.assertEqual(
                    test_data,
                    parsed_data,
                    f"Values for key '{key}' do not match.",
                )

    def test_idat_parser(self):
        test_data_list = [
            {},
            {"illumina_ids": np.array([1, 10, 1000, 10000], dtype="<i4")},
            {"illumina_ids": np.arange(0, 10000, dtype="<i4")},
            {"illumina_ids": np.arange(100, 10000, dtype="<i4")},
            {
                "n_snps_read": 20,
                "illumina_ids": np.arange(1000, 1000 + 20, dtype="<i4"),
                "probe_means": np.arange(100, 100 + 20, dtype="<u2"),
                "std_dev": np.arange(2000, 2000 + 20, dtype="<u2"),
                "n_beads": np.arange(3, 3 + 20, dtype="<u1"),
                "mid_block": np.arange(400000, 400000 + 20, dtype="<i4"),
                "run_info": [
                    [
                        "time_0",
                        "type_0",
                        "pars_0",
                        "code_0",
                        "vers_0",
                    ],
                    [
                        "time_1",
                        "type_1",
                        "pars_1",
                        "code_1",
                        "vers_1",
                    ],
                ],
                "red_green": 1,
                "mostly_null": "asdf",
                "barcode": "ghij",
                "chip_type": "klmn",
                "mostly_a": "opqr",
                "unknown_1": "stuv",
                "unknown_2": "wxxz",
                "unknown_3": "1234",
                "unknown_4": "5678",
                "unknown_5": "90.,",
                "unknown_6": "<>-:",
                "unknown_7": "!=|#",
            },
        ]

        for test_data in test_data_list:
            for gzipped in [False, True]:
                self._check_idat_parser(test_data, gzipped)


if __name__ == "__main__":
    unittest.main()



"""Module providing access to various data types and utilities."""

from .arrays import ArrayType
from .beads import (
    NEUTRAL_BETA,
    MethylData,
    RawData,
    ReferenceMethylData,
    _overlap_indices,
    idat_basepaths,
    is_valid_idat_basepath,
)
from .cache import get_id_tuple, input_args_id, memoize
from .chromosome import Chromosome
from .cnv import CNV, ZIP_ENDING, Annotation, _get_cgsegment
from .genetic_data import (
    CHROMOSOME_DATA,
    GAPS,
    GENES,
    IMPORTANT_GENES,
)
from .idat import IdatParser
from .manifests import Manifest
from .plots import (
    CNVPlot,
    cnv_plot_from_data,
    get_cn_summary,
    read_cnv_data_from_disk,
)
from .probes import Channel, InfiniumDesignType, ProbeType

__all__ = [
    "Annotation",
    "ArrayType",
    "CHROMOSOME_DATA",
    "CNV",
    "CNVPlot",
    "NEUTRAL_BETA",
    "input_args_id",
    "Channel",
    "Chromosome",
    "GAPS",
    "GENES",
    "IMPORTANT_GENES",
    "IdatParser",
    "InfiniumDesignType",
    "Manifest",
    "MethylData",
    "ProbeType",
    "RawData",
    "ReferenceMethylData",
    "ZIP_ENDING",
    "cache",
    "cnv_plot_from_data",
    "get_cn_summary",
    "get_id_tuple",
    "idat_basepaths",
    "is_valid_idat_basepath",
    "memoize",
    "_overlap_indices",
    "_get_cgsegment",
    "read_cnv_data_from_disk",
]



"""Contains methods for supervised learning.

Non supervised classifiers (random forest, k-nearest neighbors, neural
networks) for predicting the methylation class.
"""

import hashlib
import logging
import pickle
from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import (
    LinearDiscriminantAnalysis,
    QuadraticDiscriminantAnalysis,
)
from sklearn.ensemble import (
    AdaBoostClassifier,
    BaggingClassifier,
    ExtraTreesClassifier,
    GradientBoostingClassifier,
    HistGradientBoostingClassifier,
    RandomForestClassifier,
    StackingClassifier,
)
from sklearn.feature_selection import (
    SelectKBest,
    mutual_info_classif,
)
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import (
    LogisticRegression,
    Perceptron,
    RidgeClassifier,
    SGDClassifier,
)
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    precision_score,
    recall_score,
    roc_auc_score,
)
from sklearn.model_selection import (
    cross_val_predict,
)
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import (
    MinMaxScaler,
    PowerTransformer,
    QuantileTransformer,
    RobustScaler,
    StandardScaler,
)
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from mepylome.dtypes.cache import input_args_id

logger = logging.getLogger(__name__)


class TrainedClassifier(ABC):
    """Abstract base class for a trained classifier."""

    @abstractmethod
    def predict_proba(self, betas, id_=None):
        """Predicts the probability of the given input samples (betas).

        Args:
            betas (array-like): Input samples to predict probabilities.
            id_ (string, optional): Identifier of the input samples.

        Returns:
            array-like: Predicted probabilities for each sample.
        """

    @abstractmethod
    def classes(self):
        """Returns the list of classes for which predictions can be made.

        Returns:
            array-like: The classes the classifier can predict.
        """

    def info(self):
        """Provides additional information of the classifier or its pipeline.

        The info will be printed on the top of the classifiers report. Should
        contain classifier name and classifier metrics.

        Returns:
            str: A description of the classifier or its components.
        """
        return ""

    def model(self):
        """Returns the classifier model (usually sklearn object)."""
        return

    def metrics(self):
        """Returns the metric statistics (usually sklearn object)."""
        return {}

    def __str__(self):
        hash_str = hashlib.blake2b(
            self.info().encode(), digest_size=16
        ).hexdigest()
        return f"TrainedClassifier_{hash_str}"

    def __repr__(self):
        return str(self)


def _get_pipeline_description(clf):
    """Generates a detailed summary string of the pipeline structure."""

    def format_non_default_params(step):
        """Formats non-default parameters for a given step or classifier."""
        default_params = step.__class__().get_params()
        current_params = step.get_params()
        non_default_params = {
            k: v
            for k, v in current_params.items()
            if default_params.get(k) != v
        }
        return [
            f"    {param}: {value}"
            for param, value in non_default_params.items()
        ]

    lines = ["Pipeline Structure:"]
    if hasattr(clf, "steps"):
        step_name_len = max(len(name) for name, _ in clf.steps)
        for name, step in clf.steps:
            step_name = name.capitalize()
            step_type = (
                step if isinstance(step, str) else step.__class__.__name__
            )
            lines.append(f"- {step_name:<{step_name_len}} : {step_type}")
            if not isinstance(step, str):
                lines.extend(format_non_default_params(step))
    else:
        clf_name = clf.__class__.__name__
        lines.append(f"- Classifier: {clf_name}")
        lines.extend(format_non_default_params(clf))
    return "\n".join(lines)


class TrainedSklearnClassifier(TrainedClassifier):
    """Trained classifier implementation using fitted scikit-learn objecs."""

    def __init__(self, clf, X=None, metrics=None):
        self.clf = clf
        self._classes = clf.classes_
        self.ids = [] if X is None else X.index
        self._metrics = metrics or {
            "n_features": X.shape[1] if X is not None else 0,
            "n_samples": X.shape[0] if X is not None else 0,
        }

    def predict_proba(self, betas, id_=None):
        return self.clf.predict_proba(betas)

    def classes(self):
        return self._classes

    def info(self):
        description = _get_pipeline_description(self.clf)
        result = description + "\n\nMetrics:"
        formatted_stats = _format_stats(self._metrics)
        max_key_length = max(len(key) for key in formatted_stats)
        for key, value in formatted_stats.items():
            result += f"\n- {key:<{max_key_length}} : {value}"
        return result

    def model(self):
        return self.clf

    def metrics(self):
        return self._metrics


class TrainedSklearnCVClassifier(TrainedClassifier):
    """A trained sklearn classifier with support for cross-validation.

    This class allows the use of precomputed cross-validation probabilities or
    standard predictions from a trained classifier for a given sample. It also
    provides additional statistics about the classifier.
    """

    def __init__(self, clf, probabilities_cv, X, metrics=None):
        self.clf = clf
        self.probabilities_cv = pd.DataFrame(probabilities_cv, index=X.index)
        self.ids = X.index
        self._metrics = metrics or {}
        self._classes = clf.classes_

    def predict_proba(self, betas, id_=None):
        if id_ is not None:
            id_ = np.ravel(id_)
            probabilities = np.zeros((len(betas), len(self._classes)))

            if len(betas) != len(id_):
                msg = "Length of 'betas' and 'id_' must match."
                raise ValueError(msg)

            mask_known_ids = np.isin(id_, self.ids)
            if any(mask_known_ids):
                probabilities[mask_known_ids] = self.probabilities_cv.loc[
                    id_[mask_known_ids]
                ].values
            unknown_betas = betas[~mask_known_ids]
            if len(unknown_betas) > 0:
                probabilities[~mask_known_ids] = self.clf.predict_proba(
                    unknown_betas
                )
            return probabilities

        return self.clf.predict_proba(betas)

    def classes(self):
        return self._classes

    def info(self):
        description = _get_pipeline_description(self.clf)
        result = description + "\n\nMetrics:"
        formatted_stats = _format_stats(self._metrics)
        max_key_length = max(len(key) for key in formatted_stats)
        for key, value in formatted_stats.items():
            result += f"\n- {key:<{max_key_length}} : {value}"
        return result

    def metrics(self):
        return self._metrics

    def model(self):
        return self.clf


def make_clf_pipeline(scaler, selector, clf, X_shape, cv):
    """Sklearn pipeline with scaling, feature selection, and classifier."""
    n_splits = cv if isinstance(cv, int) else cv.n_splits
    n_components_pca = min(((n_splits - 1) * X_shape[0] // n_splits), 50)
    scalers = {
        "minmax": MinMaxScaler(),
        "none": "passthrough",
        "power": PowerTransformer(),
        "quantile": QuantileTransformer(output_distribution="uniform"),
        "quantile_normal": QuantileTransformer(output_distribution="normal"),
        "robust": RobustScaler(),
        "std": StandardScaler(),
    }
    selectors = {
        "kbest": SelectKBest(k=10000),
        "lda": LinearDiscriminantAnalysis(n_components=1),
        "mutual_info": SelectKBest(mutual_info_classif, k=10000),
        "none": "passthrough",
        "pca": PCA(n_components=n_components_pca),
    }
    classifiers = {
        "ada": AdaBoostClassifier(),
        "bag": BaggingClassifier(),
        "dt": DecisionTreeClassifier(),
        "et": ExtraTreesClassifier(),
        "gb": GradientBoostingClassifier(),
        "gp": GaussianProcessClassifier(),
        "hgb": HistGradientBoostingClassifier(),
        "knn": KNeighborsClassifier(n_neighbors=5, weights="distance"),
        "lda": LinearDiscriminantAnalysis(),
        "lr": LogisticRegression(max_iter=10000),
        "mlp": MLPClassifier(verbose=True),
        "nb": GaussianNB(),
        "none": "passthrough",
        "perceptron": Perceptron(max_iter=1000, tol=1e-3),
        "qda": QuadraticDiscriminantAnalysis(),
        "rf": RandomForestClassifier(),
        "ridge": RidgeClassifier(),
        "sgd": SGDClassifier(max_iter=1000, tol=1e-3),
        "stacking": StackingClassifier(
            estimators=[
                ("rf", RandomForestClassifier()),
                ("svc", SVC(kernel="linear", probability=True)),
            ],
            final_estimator=LogisticRegression(),
        ),
        "svc_linear": SVC(kernel="linear", probability=True, verbose=True),
        "svc_rbf": SVC(kernel="rbf", probability=True, verbose=True),
    }
    scaler = scalers[scaler]
    selector = selectors[selector]
    classifier = classifiers[clf]
    return Pipeline(
        [
            ("scaler", scaler),
            ("feature_selection", selector),
            ("classifier", classifier),
        ]
    )


def _make_reports(prediction, info):
    """Generates detailed reports from classifier predictions.

    Args:
        prediction (pd.DataFrame): DataFrame containing predicted probabilities
            for each class, indexed by sample IDs.
        info (str): A description of the classifier such as its name and
            metrics. Will be printed before predictions.

    Returns:
        list[str]: A list of detailed string reports, one for each sample.
    """
    reports = []

    for sample_id, row in prediction.iterrows():
        top_predictions = row[row > 0].nlargest(10) * 100
        report_lines = [
            str(sample_id),
            "=" * len(str(sample_id)),
            f"\n{info}\n",
            "Classification Probability:",
        ]
        label_len = max(len(str(label)) for label in top_predictions.index)
        formatted_lines = [
            f"- {label:<{label_len}} : {probability:6.2f} %"
            for label, probability in top_predictions.items()
        ]
        n_dashes = (
            max(len(line) for line in formatted_lines)
            if formatted_lines
            else 0
        )
        dashes = "-" * n_dashes
        report_lines.extend([dashes, *formatted_lines, dashes])
        reports.append("\n".join(report_lines))
    return reports


def _is_ovr_or_ovo(clf):
    """Check if classifier is using One-vs-Rest (OvR) or One-vs-One (OvO)."""
    if isinstance(clf, OneVsRestClassifier):
        return "ovr"
    if isinstance(clf, OneVsOneClassifier):
        return "ovo"
    if hasattr(clf, "decision_function_shape"):
        if clf.decision_function_shape == "ovr":
            return "ovr"
        if clf.decision_function_shape == "ovo":
            return "ovo"
    # Default for most classifiers
    return "ovr"


def cross_val_metrics(clf, X, y, probabilities_cv, cv):
    """Calculates cross-validation statistics for a classifier."""
    y_pred_cv = np.array(
        [clf.classes_[i] for i in np.argmax(probabilities_cv, axis=1)]
    )
    accuracy_scores = []
    auc_scores = []
    f1_scores = []
    multi_class_strategy = _is_ovr_or_ovo(clf)
    precision_scores = []
    recall_scores = []

    for _, test_idx in cv.split(X, y):
        y_true = np.array(y)[test_idx]
        probabilities_cv_test = probabilities_cv[test_idx, :]
        y_pred = y_pred_cv[test_idx]

        accuracy_scores.append(accuracy_score(y_true, y_pred))
        f1_scores.append(f1_score(y_true, y_pred, average="weighted"))
        precision_scores.append(
            precision_score(
                y_true, y_pred, average="weighted", zero_division=0
            )
        )
        recall_scores.append(
            recall_score(y_true, y_pred, average="weighted", zero_division=0)
        )

        # Calculate ROC AUC score
        if len(np.unique(y_true)) == 2:
            # Binary classification: use probabilities for the positive class
            roc_auc = roc_auc_score(y_true, probabilities_cv_test[:, 1])
        else:
            # Multiclass classification: use the full probabilities matrix
            roc_auc = roc_auc_score(
                y_true,
                probabilities_cv_test,
                multi_class=multi_class_strategy,
                average="weighted",
            )
        auc_scores.append(roc_auc)

    return {
        "cv": cv,
        "n_samples": X.shape[0],
        "n_features": X.shape[1],
        "accuracy_scores": accuracy_scores,
        "auc_scores": auc_scores,
        "f1_scores": f1_scores,
        "precision_scores": precision_scores,
        "recall_scores": recall_scores,
    }


def _format_stats(metrics):
    """Transforms classifier metrics to readable text."""

    def format_metric_scores(scores):
        if isinstance(scores, (list, np.ndarray)):
            return f"{np.mean(scores):.4f} (SD {np.std(scores):.4f})"
        return str(scores)

    formatted_stats = {}

    metric_keys = {
        "Method": "cv",
        "Samples": "n_samples",
        "Features": "n_features",
        "Accuracy": "accuracy_scores",
        "AUC": "auc_scores",
        "F1-Score": "f1_scores",
        "Precision": "precision_scores",
        "Recall": "recall_scores",
    }

    for key, metric in metric_keys.items():
        if metric in metrics:
            if metric == "cv":
                formatted_stats[key] = (
                    f"{metrics[metric].n_splits}-fold cross validation"
                )
            else:
                formatted_stats[key] = format_metric_scores(metrics[metric])

    return formatted_stats


def is_trained(clf):
    """Checks if the given classifier has already been trained."""
    clf_ = clf
    if hasattr(clf, "steps"):
        clf_ = clf.steps[-1][1]
    trained_attributes = ["classes_"]
    return all(hasattr(clf_, attr) for attr in trained_attributes)


def train_clf(clf, X, y, directory, cv, n_jobs=1):
    """Trains a classifier and stores the trained model to disk.

    If the classifier has already been trained (and saved), it loads the
    trained model. If not, it trains the classifier using the provided data and
    saves it to disk.

    Args:
        clf (classifier): The classifier to train or load.
        X (array-like): The feature matrix.
        y (array-like): The target labels.
        directory (Path): The directory where the trained model should be
            saved.
        cv (int or cross-validation generator): Determines the cross-validation
            splitting strategy.
        n_jobs (int): Number of parallel processes to run.

    Returns:
        TrainedSklearnClassifier or TrainedSklearnCVClassifier: The trained
            classifier object.
    """
    n_splits = cv if isinstance(cv, int) else cv.n_splits

    clf_filename = input_args_id(clf, cv, X.shape, len(y)) + ".pkl"

    clf_path = directory / clf_filename

    if clf_path.exists():
        with clf_path.open("rb") as file:
            return pickle.load(file)

    if is_trained(clf):
        return TrainedSklearnClassifier(clf, X=X)

    logger.info("Start training...")

    clf.fit(X, y)
    counts_per_class = np.unique(y, return_counts=True)[1]

    if min(counts_per_class) < n_splits:
        logger.info(
            "Warning: One of the classes has fewer than "
            f"{n_splits} samples (cv splits). Stats may not be computable."
        )
        trained_clf = TrainedSklearnClassifier(clf=clf, X=X)
    else:
        logger.info("Start cross-validation...")
        probabilities_cv = cross_val_predict(
            clf, X, y, cv=cv, method="predict_proba", n_jobs=n_jobs
        )
        metrics = cross_val_metrics(clf, X, y, probabilities_cv, cv)
        trained_clf = TrainedSklearnCVClassifier(
            clf=clf, probabilities_cv=probabilities_cv, X=X, metrics=metrics
        )

    with clf_path.open("wb") as file:
        pickle.dump(trained_clf, file)

    return trained_clf


@dataclass
class ClassifierResult:
    """Data container for evaluation of classifier."""

    prediction: any
    model: any
    metrics: dict
    reports: list


def fit_and_evaluate_clf(X, y, X_test, id_test, directory, clf, cv, n_jobs=1):
    """Predicts the methylation class by supervised learning classifier.

    Uses supervised machine learning classifiers (Random Forest, K-Nearest
    Neighbors, Neural Networks, SVM, ...) to predict the methylation class of
    the sample. Output will be written to disk.

    Args:
        X (pd.DataFrame): Feature matrix (rows as samples, columns as
            features).
        y (array-like): Class labels.
        X_test (array-like): Value of the sample to be evaluated.
        id_test (str): Unique identifiers for the test samples to be evaluated.
        directory (str or Path): Directory where the classifiers and results
            will be saved/cached.
        clf (list): Classifier to use. Can be:

            - A scikit-learn classifier object or pipeline (trained or
              untrained).
            - A string in the format "scaler-selector-classifier". Possible
              values are:

                scaler:
                    - "none": No scaling (passthrough).
                    - "std": Standard scaling (StandardScaler).
                    - "minmax": Min-max scaling (MinMaxScaler).
                    - "robust": Robust scaling (RobustScaler).
                    - "power": Power transformation (PowerTransformer).
                    - "quantile": Quantile transformation (QuantileTransformer,
                      uniform distribution).
                    - "quantile_normal": Quantile transformation
                      (QuantileTransformer, normal distribution).

                selector:
                    - "none": No feature selection (passthrough).
                    - "kbest": Select the best features (SelectKBest).
                    - "pca": Principal component analysis (PCA).
                    - "lda": Linear Discriminant Analysis (LDA).
                    - "mutual_info": Select features based on mutual
                      information (SelectKBest with mutual_info_classif).

                clf:
                    - "rf": RandomForestClassifier.
                    - "lr": LogisticRegression.
                    - "et": ExtraTreesClassifier.
                    - "knn": KNeighborsClassifier.
                    - "mlp": MLPClassifier.
                    - "svc_linear": Support Vector Classifier (linear kernel).
                    - "svc_rbf": Support Vector Classifier (RBF kernel).
                    - "ada": AdaBoostClassifier.
                    - "bag": BaggingClassifier.
                    - "dt": DecisionTreeClassifier.
                    - "gp": GaussianProcessClassifier.
                    - "hgb": HistGradientBoostingClassifier.
                    - "nb": GaussianNB.
                    - "perceptron": Perceptron.
                    - "qda": Quadratic Discriminant Analysis (QDA).
                    - "ridge": RidgeClassifier.
                    - "sgd": SGDClassifier.
                    - "stacking": StackingClassifier (combines multiple
                      classifiers).
                    - "none": No classifier (passthrough).

            - A custom class, that inherits from `TrainedClassifier`.

        cv (int or cross-validation generator): Determines the cross-validation
            splitting strategy.
        n_jobs (int): Number of parallel processes to run.

    Returns:
        ClassifierResult:
            - prediction (DataFrame): DataFrame containing the predicted
              probabilities for each class.
            - model (object): The trained classifier object.
            - metrics (dict): Dict containing classifier metrics.
            - reports (list): List of evaluation report strings for each
              sample.
    """
    if isinstance(clf, (Pipeline, ClassifierMixin)):
        trained_clf = train_clf(
            clf=clf, X=X, y=y, directory=directory, cv=cv, n_jobs=n_jobs
        )
    elif isinstance(clf, str):
        pipeline = make_clf_pipeline(*clf.split("-"), X.shape, cv)
        trained_clf = train_clf(
            clf=pipeline, X=X, y=y, directory=directory, cv=cv, n_jobs=n_jobs
        )
    else:
        trained_clf = clf
    classes = trained_clf.classes()
    probabilities = trained_clf.predict_proba(X_test, id_test)
    info = trained_clf.info()
    prediction = pd.DataFrame(probabilities, index=id_test, columns=classes)
    reports = _make_reports(prediction, info)
    return ClassifierResult(
        prediction,
        trained_clf.model(),
        trained_clf.metrics(),
        reports,
    )



"""Contains auxiliary functions and a debug timer.

Usage:
    timer = Timer()
    timer.start()
    # process to be measured
    timer.stop("process_name")
"""

import logging
import socket
import tempfile
import time
from datetime import datetime
from pathlib import Path
from uuid import uuid4

import numpy as np

__all__ = ["Timer", "normexp_get_xs", "MEPYLOME_TMP_DIR"]

logger = logging.getLogger(__name__)

MEPYLOME_TMP_DIR = Path(tempfile.gettempdir()) / "mepylome"
LOG_DIR = MEPYLOME_TMP_DIR / "log"

LOG_DIR.mkdir(parents=True, exist_ok=True)


def make_log_file(suffix):
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    unique_id = uuid4().hex[:8]
    log_file = LOG_DIR / f"{suffix}-{timestamp}-{unique_id}.log"
    log_file.touch(exist_ok=True)
    return log_file


class Timer:
    """Measures the time elapsed in milliseconds."""

    def __init__(self):
        self.time0 = time.time()

    def start(self):
        """Resets timer."""
        self.time0 = time.time()

    def stop(self, text=None):
        """Resets timer and return elapsed time."""
        delta_time = 1000 * (time.time() - self.time0)
        comment = "" if text is None else "(" + text + ")"
        print("Time passed:", delta_time, "ms", comment)
        self.time0 = time.time()
        return delta_time


def huber(y, k=1.5, tol=1.0e-6):
    """Compute Huber's M-estimator of location with MAD scale.

    Code adapted from
    https://github.com/cran/MASS/blob/master/R/huber.R

    Args:
        y (array-like): Vector of data values.
        k (float, optional): Winsorizes at `k` standard deviations. Default is
            1.5.
        tol (float, optional): Convergence tolerance. Default is 1.0e-6.

    Returns:
        tuple: A tuple containing:
            - mu (float): Location estimate (mean).
            - s (float): MAD scale estimate (standard deviation).

    Raises:
        ValueError: If the scale (MAD) is zero for the given sample.

    References:
        Huber, P. J. (1981). _Robust Statistics._ Wiley.
        Venables, W. N. and Ripley, B. D. (2002). _Modern Applied Statistics
            with S._ Fourth edition. Springer.

    Examples:
        >>> huber(chem)
    """
    y = y[~np.isnan(y)]
    mu = np.median(y)
    s = np.median(np.abs(y - mu)) * 1.4826
    if s == 0:
        msg = "Cannot estimate scale: MAD is zero for this sample"
        raise ValueError(msg)
    while True:
        yy = np.clip(y, mu - k * s, mu + k * s)
        mu1 = np.mean(yy)
        if np.abs(mu - mu1) < tol * s:
            break
        mu = mu1
    return mu, s


def normexp_signal(par, x):
    """Expected Signal Given Observed Foreground Under Normal+Exp Model.

    Adjust foreground intensities for observed background using Normal+Exp
    Model.

    Code adapted from:
    https://github.com/gangwug/limma/blob/master/R/background-normexp.R

    Args:
        par (array-like): Numeric vector containing the parameters of the
            Normal+Exp distribution.
                - par[0]: mu
                - par[1]: log sigma
                - par[2]: log alpha
        x (array-like): Numeric vector of (background corrected) intensities.

    Returns:
        array-like: Numeric vector containing adjusted intensities.

    Raises:
        ValueError: If alpha or sigma are non-positive.

    References:
        Ritchie, M. E., Silver, J., Oshlack, A., Silver, J., Holmes, M.,
        Diyagama, D., Holloway, A., and Smyth, G. K. (2007). A comparison of
        background correction methods for two-colour microarrays.
        _Bioinformatics_ 23, 2700-2707.
        <http://bioinformatics.oxfordjournals.org/content/23/20/2700>

        Silver, JD, Ritchie, ME, and Smyth, GK (2009). Microarray background
        correction: maximum likelihood estimation for the normal-exponential
        convolution. _Biostatistics_ 10, 352-363.
        <http://biostatistics.oxfordjournals.org/content/10/2/352>

    Examples:
        >>> normexp_signal([1, np.log(2), np.log(3)], 4)
        2.3735035872302235
    """
    from scipy.stats import norm

    mu = par[0]
    sigma = np.exp(par[1])
    sigma2 = sigma * sigma
    alpha = np.exp(par[2])
    if alpha <= 0:
        msg = "alpha must be positive"
        raise ValueError(msg)
    if sigma <= 0:
        msg = "sigma must be positive"
        raise ValueError(msg)
    mu_sf = x - mu - sigma2 / alpha
    log_dnorm = norm.logpdf(0, loc=mu_sf, scale=sigma)
    log_pnorm = norm.logsf(0, loc=mu_sf, scale=sigma)
    signal = mu_sf + sigma2 * np.exp(log_dnorm - log_pnorm)
    z = ~np.isnan(signal)
    if np.any(signal[z] < 0):
        logger.warning(
            "Limit of numerical accuracy reached with very low intensity or "
            "very high background:\nsetting adjusted intensities to small "
            "value"
        )
        signal[z] = np.maximum(signal[z], 1e-6)
    return signal


def normexp_get_xs(xf, controls=None, offset=50, param=None):
    """Used in NOOB.

    Adapted from
    https://github.com/hansenlab/minfi/blob/devel/R/preprocessNoob.R
    """
    n_probes = xf.shape[0]
    if param is None:
        if controls is None:
            msg = "'controls' or 'param' must be given"
            raise ValueError(msg)
        alpha = np.empty(n_probes)
        mu = np.empty(n_probes)
        sigma = np.empty(n_probes)
        for i in range(n_probes):
            mu[i], sigma[i] = huber(controls[i, :])
            alpha[i] = max(huber(xf[i, :])[0] - mu[i], 10)
        param = np.column_stack((mu, np.log(sigma), np.log(alpha)))
    result = np.empty(xf.shape)
    for i in range(n_probes):
        result[i, :] = normexp_signal(param[i], xf[i, :])
    return {
        "xs": result + offset,
        "param": param,
    }


def get_free_port(start_port):
    """Returns the first free port from start position."""
    port = start_port
    while True:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex(("localhost", port)) != 0:
                return port
            port += 1



"""Defines an Enum representing chromosomes with utility methods.

Usage:
    chr_2 = Chromosome.CHR2
    chr_x = Chromosome.CHRX

    is_valid = Chromosome.is_valid_chromosome(chr_x)
    df = pd.DataFrame({'Chromosome_0': ["13", "chr1", "chr2", "chrY", "chrx"]})
    df['Chromosome_1'] = Chromosome.pd_from_string(df['Chromosome_0'])
    df['Chromosome_2'] = Chromosome.pd_to_string(df['Chromosome_1'])
"""

from enum import IntEnum, unique


@unique
class Chromosome(IntEnum):
    """Enum representing chromosomes and providing utility methods.

    Chromosomes are represented as integers:
    - 0-22: Chromosomes 1-22
    - 23: Chromosome X
    - 24: Chromosome Y
    - 25: Mitochondrial Chromosome (MT)
    - -1: Invalid Chromosome

    Methods:
        - is_valid_chromosome(chrom): Checks if the given chromosome is valid.
        - pd_from_string(col): Converts chromosome strings to Chromosome enum
            values.
        - pd_to_string(col): Converts Chromosome enum values to chromosome
            strings.
    """

    CHR0 = 0
    CHR1 = 1
    CHR2 = 2
    CHR3 = 3
    CHR4 = 4
    CHR5 = 5
    CHR6 = 6
    CHR7 = 7
    CHR8 = 8
    CHR9 = 9
    CHR10 = 10
    CHR11 = 11
    CHR12 = 12
    CHR13 = 13
    CHR14 = 14
    CHR15 = 15
    CHR16 = 16
    CHR17 = 17
    CHR18 = 18
    CHR19 = 19
    CHR20 = 20
    CHR21 = 21
    CHR22 = 22
    CHRX = 23
    CHRY = 24
    CHRM = 25
    INVALID = -1

    @staticmethod
    def is_valid_chromosome(chrom):
        """Checks if the given chromosome is valid."""
        return (chrom >= Chromosome.CHR1) & (chrom <= Chromosome.CHRY)

    @staticmethod
    def pd_from_string(col):
        """Converts chromosome strings to Chromosome enum values."""
        chrom_map = {
            **{str(i): Chromosome(i) for i in range(23)},
            **{"chr" + str(i): Chromosome(i) for i in range(23)},
            "X": Chromosome.CHRX,
            "Y": Chromosome.CHRY,
            "M": Chromosome.CHRM,
            "x": Chromosome.CHRX,
            "y": Chromosome.CHRY,
            "m": Chromosome.CHRM,
            "chrX": Chromosome.CHRX,
            "chrY": Chromosome.CHRY,
            "chrM": Chromosome.CHRM,
            "chrx": Chromosome.CHRX,
            "chry": Chromosome.CHRY,
            "chrm": Chromosome.CHRM,
        }
        return col.map(chrom_map).fillna(Chromosome.INVALID).astype(int)

    @staticmethod
    def pd_to_string(col):
        """Converts Chromosome enum values to chromosome strings."""
        chrom_map = {
            **{Chromosome(i): "chr" + str(i) for i in range(23)},
            Chromosome.CHRX: "chrX",
            Chromosome.CHRY: "chrY",
            Chromosome.CHRM: "chrM",
        }
        return col.map(chrom_map).fillna("NaN")



"""Contains a IDAT file parser."""

import gzip
import io
import os
from enum import IntEnum, unique

import numpy as np

from mepylome.utils.files import get_file_object

__all__ = ["IdatParser"]

DEFAULT_IDAT_VERSION = 3
DEFAULT_IDAT_FILE_ID = "IDAT"


def read_byte(infile):
    return int.from_bytes(infile.read(1), byteorder="little", signed=False)


def read_short(infile):
    return int.from_bytes(infile.read(2), byteorder="little", signed=False)


def read_int(infile):
    return int.from_bytes(infile.read(4), byteorder="little", signed=True)


def read_long(infile):
    return int.from_bytes(infile.read(8), byteorder="little", signed=True)


def read_char(infile, num_bytes):
    return infile.read(num_bytes).decode("utf-8")


def read_string(infile):
    num_bytes = read_byte(infile)
    num_chars = num_bytes % 128
    shift = 0
    while num_bytes // 128 == 1:
        num_bytes = read_byte(infile)
        shift += 7
        offset = (num_bytes % 128) * (2**shift)
        num_chars += offset
    return read_char(infile, num_chars)


def read_array(infile, dtype, n):
    dtype = np.dtype(dtype)
    total_size = dtype.itemsize * n
    alldata = infile.read(total_size)

    if len(alldata) != total_size:
        msg = "End of file reached before number of results parsed"
        raise EOFError(msg)

    return np.frombuffer(alldata, dtype)


@unique
class IdatSectionCode(IntEnum):
    """Section codes used in Illumina idat files."""

    ILLUMINA_ID = 102
    STD_DEV = 103
    MEAN = 104
    NUM_BEADS = 107  # how many replicate measurements for each probe
    MID_BLOCK = 200
    RUN_INFO = 300
    RED_GREEN = 400
    MOSTLY_NULL = 401  # manifest
    BARCODE = 402
    CHIP_TYPE = 403  # format
    MOSTLY_A = 404  # label
    UNKNOWN_1 = 405  # opa
    UNKNOWN_2 = 406  # sampleid
    UNKNOWN_3 = 407  # descr
    UNKNOWN_4 = 408  # plate
    UNKNOWN_5 = 409  # well
    UNKNOWN_6 = 410
    UNKNOWN_7 = 510  # unknown
    NUM_SNPS_READ = 1000


def _get_file_size(file_like):
    """Get the size of a file-like object."""
    # Check if the file-like object has a fileno method
    if isinstance(file_like, (io.BufferedReader, gzip.GzipFile)):
        return os.fstat(file_like.fileno()).st_size

    if isinstance(file_like, io.BytesIO):
        current_pos = file_like.tell()
        file_like.seek(0, io.SEEK_END)
        size = file_like.tell()
        file_like.seek(current_pos)
        return size

    msg = "Cannot determine file size. Unknown file format"
    raise ValueError(msg)


class IdatParser:
    """Reads and parses an IDAT file.

    Stores all extracted values from the IDAT file as attributes.

    Args:
        file (str or file-like object): Path to the IDAT file or
            a file-like object. Can also be a gzipped IDAT file.
        intensity_only (bool, optional): Whether to read only intensity values,
            which makes parsing faster. Defaults to False.

    Examples:
        >>> filepath = "/path/to/idat/file_Grn.idat"
        >>> idat_data = IdatParser(filepath)
        >>> ids = idat_data.illumina_ids
        >>> print(idat_data)
    """

    def __init__(
        self,
        file,
        *,
        intensity_only=False,
        array_type_only=False,
    ):
        """Reads and parses the IDAT file."""
        self.intensity_only = intensity_only
        self.array_type_only = array_type_only
        self._file = file

        with get_file_object(file) as idat_file:
            self.file_size = _get_file_size(idat_file)
            self._parse_header(idat_file)
            self._parse_body(idat_file)

    def _parse_header(self, idat_file):
        file_type = read_char(idat_file, len(DEFAULT_IDAT_FILE_ID))
        # Assert file is indeed IDAT format
        if file_type != DEFAULT_IDAT_FILE_ID:
            msg = (
                f"Parser could not open file {self._file} as its not a valid "
                "IDAT file."
            )
            raise ValueError(msg)

        idat_version = read_long(idat_file)

        # Assert correct IDAT file version
        if idat_version != DEFAULT_IDAT_VERSION:
            msg = (
                f"Parser could not open file {self._file} as its not a "
                "version 3 IDAT file."
            )
            raise ValueError(msg)

        self.num_fields = read_int(idat_file)

        self.offsets = {
            read_short(idat_file): read_long(idat_file)
            for _ in range(self.num_fields)
        }

    def _parse_body(self, idat_file):
        def seek_to_section(section_code):
            idat_file.seek(self.offsets[section_code.value])

        seek_to_section(IdatSectionCode.NUM_SNPS_READ)
        self.n_snps_read = read_int(idat_file)

        if self.array_type_only:
            return

        seek_to_section(IdatSectionCode.ILLUMINA_ID)
        self.illumina_ids = read_array(idat_file, "<i4", self.n_snps_read)

        seek_to_section(IdatSectionCode.MEAN)
        self.probe_means = read_array(idat_file, "<u2", self.n_snps_read)

        if self.intensity_only:
            return

        seek_to_section(IdatSectionCode.STD_DEV)
        self.std_dev = read_array(idat_file, "<u2", self.n_snps_read)

        seek_to_section(IdatSectionCode.NUM_BEADS)
        self.n_beads = read_array(idat_file, "<u1", self.n_snps_read)

        seek_to_section(IdatSectionCode.MID_BLOCK)
        n_mid_block = read_int(idat_file)
        self.mid_block = read_array(idat_file, "<i4", n_mid_block)

        seek_to_section(IdatSectionCode.RUN_INFO)
        runinfo_entry_count = read_int(idat_file)

        self.run_info = [None] * runinfo_entry_count
        for i in range(runinfo_entry_count):
            self.run_info[i] = [
                read_string(idat_file),  # run_time
                read_string(idat_file),  # block_type
                read_string(idat_file),  # block_pars
                read_string(idat_file),  # block_code
                read_string(idat_file),  # code_version
            ]

        seek_to_section(IdatSectionCode.RED_GREEN)
        self.red_green = read_int(idat_file)

        seek_to_section(IdatSectionCode.MOSTLY_NULL)
        self.mostly_null = read_string(idat_file)

        seek_to_section(IdatSectionCode.BARCODE)
        self.barcode = read_string(idat_file)

        seek_to_section(IdatSectionCode.CHIP_TYPE)
        self.chip_type = read_string(idat_file)

        seek_to_section(IdatSectionCode.MOSTLY_A)
        self.mostly_a = read_string(idat_file)

        seek_to_section(IdatSectionCode.UNKNOWN_1)
        self.unknown_1 = read_string(idat_file)

        seek_to_section(IdatSectionCode.UNKNOWN_2)
        self.unknown_2 = read_string(idat_file)

        seek_to_section(IdatSectionCode.UNKNOWN_3)
        self.unknown_3 = read_string(idat_file)

        seek_to_section(IdatSectionCode.UNKNOWN_4)
        self.unknown_4 = read_string(idat_file)

        seek_to_section(IdatSectionCode.UNKNOWN_5)
        self.unknown_5 = read_string(idat_file)

        seek_to_section(IdatSectionCode.UNKNOWN_6)
        self.unknown_6 = read_string(idat_file)

        seek_to_section(IdatSectionCode.UNKNOWN_7)
        self.unknown_7 = read_string(idat_file)

    def __repr__(self):
        with np.printoptions(edgeitems=2):
            result = (
                f"IdatParser(\n"
                f"    file_size: {self.file_size}\n"
                f"    num_fields: {self.num_fields}\n"
                f"    illumina_ids: {self.illumina_ids!r}\n"
                f"    probe_means: {self.probe_means!r}\n"
            )

            if self.intensity_only:
                return result + ")"

            return result + (
                f"    std_dev: {self.std_dev!r}\n"
                f"    n_beads: {self.n_beads!r}\n"
                f"    mid_block: {self.mid_block!r}\n"
                f"    red_green: {self.red_green}\n"
                f"    mostly_null: {self.mostly_null}\n"
                f"    barcode: {self.barcode}\n"
                f"    chip_type: {self.chip_type}\n"
                f"    mostly_a: {self.mostly_a}\n"
                f"    unknown_1: {self.unknown_1}\n"
                f"    unknown_2: {self.unknown_2}\n"
                f"    unknown_3: {self.unknown_3}\n"
                f"    unknown_4: {self.unknown_4}\n"
                f"    unknown_5: {self.unknown_5}\n"
                f"    unknown_6: {self.unknown_6}\n"
                f"    unknown_7: {self.unknown_7}\n"
                ")"
            )



"""Utility functions for caching and memoization.

Utility functions and classes for caching and memoization to optimize
performance by storing and reusing computed results.
"""

import inspect
import re
from pathlib import Path

import numpy as np
import pandas as pd
import xxhash


def np_hash(array):
    """Generates a hashable key for a NumPy array.

    This function creates a hashable key from a NumPy array by converting it to
    bytes if the data type is not an object. If the array's data type is an
    object (used for IlmnID), it samples 57 evenly spaced elements from the
    array along with the last element and combines them into a tuple with the
    array's length to form the key.

    Args:
        array (numpy.ndarray): The input array for which to generate the
            hashable key.

    Returns:
        tuple: A hashable key representing the array.

    Note:
        If 'array' is an object, collisions are possible.

    Example:
        >>> arr = np.array([1, 2, 3, 4, 5])
        >>> np_hash(arr)
    """
    if array.dtype != np.dtype(object):
        return array.tobytes()
    full_len = len(array)
    sample_len = 57
    idx_left = [i * full_len // sample_len for i in range(sample_len)] + [-1]
    return (
        tuple(array[x] for x in idx_left),
        full_len,
    )


def pd_hash(data_frame):
    """Generates a hashable key for a pandas DataFrame."""
    return (tuple(np_hash(data_frame[col].values) for col in data_frame),)


def cache_key(*args):
    """Generates a cache key for arguments based on their type.

    Args:
        arg: The input arguments, which can be of various types such as
            ArrayType, Manifest, Path, bool, int, NoneType, str,
            RangeIndex, Index, or ndarray.

    Returns:
        The cache key for the arguments.

    Warning:
        If an argument is a numpy array of object type, the key may not be
        unique.
    """
    type_map = {
        "ArrayType": str,
        "Manifest": lambda x: x.array_type.value,
        "PosixPath": str,
        "WindowsPath": str,
        "Path": str,
        "bool": str,
        "int": str,
        "NoneType": str,
        "str": str,
        "RangeIndex": lambda x: x.values.tobytes(),
        "Index": lambda x: x.values.tobytes(),
        "ndarray": np_hash,
        "DataFrame": pd_hash,
    }
    keys = tuple(type_map.get(type(arg).__name__, id)(arg) for arg in args)
    return keys[0] if len(args) == 1 else keys


def get_id_tuple(f, args, kwargs):
    """Generates a identifier tuple for a class/function and its arguments.

    This function creates a tuple that uniquely identifies a function call
    based on the function reference, its positional arguments, and keyword
    arguments. The keyword arguments are sorted by key to ensure a consistent
    order.

    Args:
        f: The function reference.
        args: A list of positional arguments passed to the function.
        kwargs: A dictionary of keyword arguments passed to the function. The
            keyword argument 'verbose' is excluded from the identifier.

    Returns:
        tuple: A tuple representing the unique identifier for the function
            call.

    Warning:
        If arg is a numpy array of object type (used in IlmnID), the key may
        not be unique.
    """
    id_list = [cache_key(f)]
    id_list.extend(cache_key(arg) for arg in args)
    # Sort keyword arguments by key to ensure consistent order
    sorted_kwargs = sorted((k, v) for k, v in kwargs.items() if k != "verbose")
    id_list.extend((key, cache_key(value)) for key, value in sorted_kwargs)
    return tuple(id_list)


def memoize(f):
    """Memoization decorator for classes and functions.

    Description:
        The `memoize` function is a decorator that provides memoization for
        class instantiation and functions. It caches instances of the decorated
        class or function based on the arguments provided. If an instance or
        function result with the same arguments already exists in the cache, it
        returns the cached instance or result instead of creating a new one.

    Args:
        f (Union[class, function]): The class or function to be decorated with
        memoization.

    Returns:
        Memoize: A memoized version of the input class or function.

    Note:
        Adapted from:
        https://stackoverflow.com/questions/10879137/how-can-i-memoize-a-class-instantiation-in-python
    """

    class Memoize:
        def __init__(self, cls):
            self.cls = cls
            self._cache = {}
            self.__name__ = cls.__name__
            self.__doc__ = cls.__doc__
            self.__module__ = cls.__module__
            init_signature = inspect.signature(cls.__init__)
            self.init_defaults = {
                key: val.default
                for key, val in init_signature.parameters.items()
                if val.default is not inspect.Parameter.empty
            }
            self.__dict__.update(cls.__dict__)
            # This bit allows staticmethods to work as you would expect.
            for attr_name, attr_value in cls.__dict__.items():
                if isinstance(attr_value, staticmethod):
                    self.__dict__[attr_name] = attr_value.__func__

        def __call__(self, *args, **kwargs):
            """Generate key."""
            complete_kwargs = {**self.init_defaults, **kwargs}
            key = get_id_tuple(self.cls, args, complete_kwargs)
            if key not in self._cache:
                self._cache[key] = self.cls(*args, **kwargs)
            return self._cache[key]

        def __instancecheck__(self, other):
            """Make isinstance() work."""
            return isinstance(other, self.cls)

    return Memoize(f)


def input_args_id(*args, extra_hash=None, suffix_limit=40):
    """Returns a unique identifier for a set of arguments."""
    components = []
    hasher = xxhash.xxh64()

    def _encode_arg(arg):
        if isinstance(arg, np.ndarray):
            return arg.tobytes()
        if isinstance(arg, pd.DataFrame):
            if not arg.select_dtypes(include=["object"]).empty:
                msg = "DataFrame contains columns with object dtype."
                raise ValueError(msg)
            return arg.values.tobytes()
        if isinstance(arg, Path):
            components.append(arg.name)
            return str(arg).encode()
        if isinstance(arg, (list, tuple)):
            return ",".join(map(str, arg)).encode()
        if hasattr(arg, "steps"):
            value = "-".join(str(x[1])[:15] for x in arg.steps)
            components.append(value)
            return str(arg).encode()
        components.append(str(arg))
        return str(arg).encode()

    for arg in args:
        hasher.update(_encode_arg(arg))
    if extra_hash:
        hasher.update(",".join(map(str, extra_hash)).encode())
    arg_hash = hasher.hexdigest()
    suffix = "-".join(components)[:suffix_limit]
    filename = f"{suffix}-{arg_hash}" if suffix else arg_hash
    return re.sub(r"[^a-zA-Z0-9_-]", "", filename)



import os
from pathlib import Path

import numpy as np
import pybind11
from setuptools import Extension, find_packages, setup

include_dirs = [
    pybind11.get_include(),
    np.get_include(),
]

# If C++ parser should be added, do: export MEPYLOME_CPP=1
add_cpp = os.getenv("MEPYLOME_CPP") == "1"

# If debug compiler flags should be used, do: export MEPYLOME_DEBUG=1
debug = os.getenv("MEPYLOME_DEBUG") == "1"

# Set the compiler flags based on the debug flag
if debug:
    print("\n*** Debugging compiler flags are used. ***\n")
    compile_args = [
        "-O0",
        "-Wall",
        "-Wextra",
        "-fno-stack-protector",
        "-g",
        "-pg",
    ]
else:
    compile_args = [
        "-O3",
        "-march=native",
    ]

sources = list(map(str, Path("pybindings").glob("*.cpp"))) + list(
    map(str, Path("src").glob("*.cpp"))
)
cpp_extension = Extension(
    name="_mepylome",
    sources=sources,
    include_dirs=include_dirs,
    extra_compile_args=compile_args,
    extra_link_args=["-fopenmp"],
    language="c++",
)

if add_cpp:
    ext_modules = [cpp_extension] if add_cpp else []
    print("\n*** Adding cpp extension. ***\n")
else:
    ext_modules = []

with open("README.md") as f:
    long_description = f.read()

setup(
    name="mepylome",
    version="0.7.8",
    description="Python package for processing Infinum DNA methylation arrays",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "numpy",
        "odfpy",
        "pandas",
        "openpyxl",
        "tqdm",
        "pyranges",
        "psutil",
        "xlrd",
        "scikit-learn",
        "dash>=2.16.0",
        "umap-learn",
        "dash_bootstrap_components",
        "importlib_resources; python_version < '3.9'",
        "xxhash",
    ],
    ext_modules=ext_modules,
    keywords="Illumina, Methylation, Infinum, Microarray, BeadChip",
    license="GPL-3.0 license",
    author="Jon Brugger",
    url="https://github.com/brj0/mepylome",
    # Include package data such as csv-Files, images, ...
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],
    entry_points={
        "console_scripts": [
            "mepylome=mepylome.analysis.cli:start_mepylome",
        ]
    },
)



"""Methylation analysis tools module.

Provides a class for performing methylation analysis and a Dash-based browser
application.
"""

from .methyl import MethylAnalysis
from .methyl_clf import TrainedClassifier

__all__ = ["MethylAnalysis", "TrainedClassifier"]
